#!/usr/bin/env python

"""
Copyright 2019 Pystol (pystol.org).

Licensed under the Apache License, Version 2.0 (the "License"); you may
not use this file except in compliance with the License. You may obtain
a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""

from pkg_resources import get_distribution

__version__ = get_distribution('pystol').version



#!/usr/bin/env python

"""
Copyright 2019 Pystol (pystol.org).

Licensed under the Apache License, Version 2.0 (the "License"); you may
not use this file except in compliance with the License. You may obtain
a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""

import json
import shutil
import tempfile
import textwrap
import urllib.request

import kubernetes
from kubernetes.client.rest import ApiException

from prettytable import PrettyTable

from pystol import __version__
from pystol.const import PYSTOL_BRANCH
from pystol.operator import load_kubernetes_config

from rich.console import Console
from rich.markdown import Markdown

pystol_version = __version__


def show_action(name):
    """
    Show the docs for specific Pystol action from Galaxy.

    This is a main component of the input for the controller
    """
    url = ("https://raw.githubusercontent.com/pystol/"
           "pystol-galaxy/" + PYSTOL_BRANCH + "/actions/roles/" +
           name + "/README.md")

    console = Console()
    try:
        with urllib.request.urlopen(url) as response:
            with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
                shutil.copyfileobj(response, tmp_file)
        with open(tmp_file.name) as html:
            markdown = Markdown(html.read())
        console.print(markdown)
    except Exception:
        print("Action not found")


def show_actions():
    """
    Show the available Pystol actions from Galaxy.

    This is a main component of the input for the controller
    """
    url = ("https://galaxy.ansible.com/api/internal/"
           "ui/repo-or-collection-detail/"
           "?namespace=pystol&name=actions")

    try:
        response = urllib.request.urlopen(url)
        data = response.read()
        values = json.loads(data)
        actions = (values['data']['collection']['latest_version']
                   ['contents'])
        description = (values['data']['collection']['latest_version']
                       ['metadata']['description'])
        repository = (values['data']['collection']['latest_version']
                      ['metadata']['repository'])
        documentation = (values['data']['collection']['latest_version']
                         ['metadata']['documentation'])
        license = (values['data']['collection']['latest_version']
                   ['metadata']['license'])
        version = (values['data']['collection']['latest_version']
                   ['metadata']['version'])
    except Exception:
        actions = {}
        description = ""
        repository = ""
        documentation = ""
        license = ""
        version = ""
        print("No objects found...")

    x = PrettyTable()
    x.title = description
    x.field_names = ["Name",
                     "Description",
                     "Documentation"]

    url = ("https://github.com/pystol/pystol-galaxy/tree/" +
           PYSTOL_BRANCH + "/actions/roles/")

    ret = []
    for action in actions:
        if action['content_type'] == "role":
            ret.append({'name': action['name'],
                        'description': action['description'],
                        'documentation': url + action['name'],
                        'license': str(license),
                        'version': str(version),
                        'repository': str(repository)})

            x.add_row([action['name'],
                       action['description'],
                       url + action['name']])
    print(x)
    print("Actions part of the Pystol core from: " + str(repository))
    print("Documentation: " + str(documentation))
    print("License: " + str(license))
    print("Version: " + str(version))
    print("Published at: https://galaxy.ansible.com/pystol")
    print("For more information go to: https://docs.pystol.org")

    return ret

def list_actions(api_client=None, debug=False):
    """
    List Pystol actions from the cluster.

    This is a main component of the input for the controller
    """
    load_kubernetes_config()
    api = kubernetes.client.CustomObjectsApi(api_client=api_client)

    group = "pystol.org"
    version = "v1alpha1"
    namespace = "pystol"
    plural = "pystolactions"
    pretty = 'true'

    x = PrettyTable()
    x.field_names = ["Name",
                     "Creation",
                     "Action state",
                     "Workflow state"]
    ret = []
    try:
        resp = api.list_namespaced_custom_object(group=group,
                                                 version=version,
                                                 namespace=namespace,
                                                 plural=plural,
                                                 pretty=pretty)
        for action in resp['items']:

            name = action['metadata']['name']
            job_description = ""
            pod_logs = ""

            if debug:

                namespace = "pystol"
                pretty = 'true'

                apiO = kubernetes.client.BatchV1Api(api_client=api_client)

                try:
                    resp = apiO.read_namespaced_job(name=name,
                                                   namespace=namespace,
                                                   pretty=pretty)
                    job_description = resp
                except ApiException:
                    print("Job not found")

                apiO = kubernetes.client.CoreV1Api(api_client=api_client)

                try:
                    resp = apiO.list_namespaced_pod(namespace=namespace,
                                                   pretty=pretty)
                    found = False
                    for pod in resp.items:
                        if name in pod.metadata.name:
                            found = True
                            resp = apiO.read_namespaced_pod_log(name=pod.metadata.name,
                                                               namespace=namespace,
                                                               pretty=pretty)
                            pod_logs = resp
                    if not found:
                        print("Pod not found.")
                except ApiException:
                    print("Pod not found.")

            ret.append({'name':
                        action['metadata']['name'],
                        'creationTimestamp':
                        action['metadata']['creationTimestamp'],
                        'action_state':
                        action['spec']['action_state'],
                        'workflow_state':
                        action['spec']['workflow_state'],
                        'stdout':
                        action['spec']['action_stdout'],
                        'stderr':
                        action['spec']['action_stderr'],
                        'job_description': job_description,
                        'pod_logs': pod_logs})

            x.add_row([action['metadata']['name'],
                       action['metadata']['creationTimestamp'],
                       action['spec']['action_state'],
                       action['spec']['workflow_state']])
    except Exception as e:
        print("No objects found or error...")
        return []
    print(x)
    return ret


def get_action(name, debug=False, api_client=None):
    """
    Get Pystol action details.

    This is a main component of the input for the controller
    """
    load_kubernetes_config()
    api = kubernetes.client.CustomObjectsApi(api_client=api_client)

    group = "pystol.org"
    version = "v1alpha1"
    namespace = "pystol"
    plural = "pystolactions"
    try:
        resp = api.get_namespaced_custom_object(group=group,
                                                version=version,
                                                namespace=namespace,
                                                plural=plural,
                                                name=name)
        print("  " + u"\U0001F4AC" + " General information")
        print("    Action: " + resp['spec']['namespace'] + "." +
              resp['spec']['collection'] + "." +
              resp['spec']['role'])

        print("    Extra variables: " + resp['spec']['extra_vars'])

        print("    Source: " + resp['spec']['source'])

        print("  " + u"\U0001F4AC" + " Status information")

        print("    Action state: " + resp['spec']['action_state'])

        print("    Workflow state: " + resp['spec']['workflow_state'])

        prefix = "    Std output: "
        preferredwidth = 70
        wrapper = textwrap.TextWrapper(initial_indent=prefix,
                                       width=preferredwidth,
                                       subsequent_indent=' ' * len(prefix))
        print(wrapper.fill(resp['spec']['action_stdout']))

        prefix = "    Std error: "
        preferredwidth = 70
        wrapper = textwrap.TextWrapper(initial_indent=prefix,
                                       width=preferredwidth,
                                       subsequent_indent=' ' * len(prefix))
        print(wrapper.fill(resp['spec']['action_stderr']))

    except ApiException:
        print("  " + u"\U0001F914" +
              " Object not found, perhaps you have a typo.")

    if debug:

        api = kubernetes.client.BatchV1Api(api_client=api_client)
        namespace = "pystol"
        pretty = 'true'
        try:
            resp = api.read_namespaced_job(name=name,
                                           namespace=namespace,
                                           pretty=pretty)
            print("---- Job description begins ----")
            print(resp)
            print("---- Job description ends ----")
        except ApiException:
            print("  " + u"\U0001F914" +
                  " Job not found, perhaps you have a typo.")

        api = kubernetes.client.CoreV1Api(api_client=api_client)
        namespace = "pystol"
        pretty = 'true'
        try:
            resp = api.list_namespaced_pod(namespace=namespace,
                                           pretty=pretty)
            found = False
            for pod in resp.items:
                if name in pod.metadata.name:
                    found = True
                    resp = api.read_namespaced_pod_log(name=pod.metadata.name,
                                                       namespace=namespace,
                                                       pretty=pretty)
                    print("---- Pod logs begins ----")
                    print("Logs from: " + pod.metadata.name)
                    print(resp)
                    print("---- Pod logs ends ----")
            if not found:
                print("  " + u"\U0001F914" +
                      " Pod not found, perhaps you have a typo.")
        except ApiException:
            print("  " + u"\U0001F914" +
                  " Pod not found, perhaps you have a typo.")



#!/usr/bin/env python

"""
Copyright 2019 Pystol (pystol.org).

Licensed under the Apache License, Version 2.0 (the "License"); you may
not use this file except in compliance with the License. You may obtain
a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""

import logging
import logging.handlers
import os
import threading
from argparse import ArgumentParser

from pystol import __version__
from pystol.cleaner import purge_pystol
from pystol.const import PYSTOL_LOG_FILE
from pystol.deployer import deploy_pystol
from pystol.get_banner import get_banner
from pystol.lister import get_action, list_actions, show_action, show_actions
from pystol.operator import insert_pystol_object
from pystol.operator import watch_for_pystol_objects, watch_for_pystol_timeouts

pystol_version = __version__

t1_stop = threading.Event()
t2_stop = threading.Event()

handler = logging.handlers.WatchedFileHandler(
    PYSTOL_LOG_FILE)
formatter = logging.Formatter(logging.BASIC_FORMAT)
handler.setFormatter(formatter)
root = logging.getLogger()
root.setLevel(os.environ.get("LOGLEVEL", "INFO"))
root.addHandler(handler)


def main():
    """
    Application's entry point.

    Here, application's settings are read from the command line,
    environment variables and CRD. Then, retrieving and processing
    of Kubernetes events are initiated.
    """
    parser = ArgumentParser(
        description='Pystol - CLI',
        prog='pystol'
    )

    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version='%(prog)s ' + pystol_version
    )

    parser.add_argument(
        '-b',
        '--banner',
        action='store_true',
        help='Print Pystol.org banner'
    )

    subparsers = parser.add_subparsers(title="Pystol subcommands",
                                       dest="command",
                                       help=("These are the options "
                                             "supported: \n"
                                             "The listen option will "
                                             "watch for CRD events. "
                                             "The run option will "
                                             "execute the Pystol "
                                             "actions against the cluster."))

    parser_purge = subparsers.add_parser('purge',
                                         help=("Purge any Pystol "
                                               "deployed resource."))

    parser_purge.add_argument(
        '-y',
        '--yes',
        action='store_true',
        help=("Run this to avoid having to write 'yes'"))

    parser_run = subparsers.add_parser('run', help=("CLI options to run the "
                                                    "Pystol actions."))

    parser_run.add_argument(
        '-n',
        '--namespace',
        required=True,
        type=str,
        help=("Name of the namespace to be referenced, "
              "this will be the Galaxy handler, like: "
              "pystol (from galaxy), "
              "which will be referenced as "
              "https://galaxy.ansible.com/pystol"))

    parser_run.add_argument(
        '-c',
        '--collection',
        required=True,
        type=str,
        help=("Name of the collection to be installed/executed, "
              "this can be the name of the galaxy collection like: "
              "actions (from galaxy), "
              "which will be referenced as "
              "https://galaxy.ansible.com/pystol/actions"))

    parser_run.add_argument(
        '-r',
        '--role',
        required=True,
        type=str,
        help=("Name of the role to be executed part of the "
              "namespace and collection value, "
              "for example, if the name is: "
              "kill-pods "
              "It will execute: "
              "pystol.actions.kill-pods "))

    parser_run.add_argument(
        '-s',
        '--source',
        default="galaxy.ansible.com",
        type=str,
        help=("The source URL where we will fetch the collection "
              "with the Pystol actions.\n"
              "It can be i.e. git+http://github.com/pystol/pystol-galaxy.git\n"
              "Defaults to: galaxy.ansible.com"))

    parser_run.add_argument(
        '-e',
        '--extra-vars',
        default="{}",
        type=str,
        help=("The extra vars for example:"
              "--extra-vars '{\"a\":\"b\",\"c\":[\",\"d\",\"e\"]}'\n"
              "Defaults to: ''"))

    subparsers.add_parser('listen', help=("CLI options to "
                                          "watch for CRs"))

    subparsers.add_parser('list', help=("CLI options to "
                                        "list CRs"))

    parser_get = subparsers.add_parser('get',
                                       help=("Get Pystol resource."))

    parser_get.add_argument(
        "action",
        type=str,
        help=("Specify the action to fetch details"))

    parser_show = subparsers.add_parser('show',
                                        help=("Get available Pystol actions."))

    parser_show.add_argument(
        "action",
        nargs='?',
        default='',
        type=str,
        help=("Specify the action to fetch details"))

    parser_get.add_argument(
        '-d',
        '--debug',
        action='store_true',
        help=("Add the debug flag to show extra information"))

    subparsers.add_parser('deploy',
                          help=("Install the Pystol operator "
                                "includes, deployment, "
                                "RBAC rules, and CRD"))

    args = parser.parse_args()

    # print("Pystol called with the folowing parameters")
    # print(parser.parse_args())

    if args.banner:
        print(get_banner())
        exit()

    try:
        if (args.command == 'run'):
            print(u"\U0001F52B" + " Starting a Pystol action")
            print("  " + u"\U0001F4BE" + " Inserting the custom resource"
                                         " in the cluster")
            ins = insert_pystol_object(args.namespace,
                                       args.collection,
                                       args.role,
                                       args.source,
                                       args.extra_vars)

            if ins:
                print("  " + u"\U0001F4A8" +
                      " The action was deployed OK")
                print("  " + u"\U0001F916" +
                      " Check its status for details")
                print("  " + u"\U0001F92B" + " Try using the CLI list and get"
                                             " options, 'pystol -h' helps")
            else:
                print("  " + u"\U0001F914" + " We can not add the resource,"
                                             " did you deploy Pystol?")
                print("  " + u"\U0001F440" + " Logs are stored in"
                                             " /tmp/pystol.org")

            exit()
        elif (args.command == 'listen'):
            print("We will watch for objects to process")
            try:
                t1 = threading.Thread(target=watch_for_pystol_objects,
                                      args=(t1_stop,))
                t1.start()
                t2 = threading.Thread(target=watch_for_pystol_timeouts,
                                      args=(t2_stop,))
                t2.start()
            except Exception as err:
                print("Error: unable to start thread: " + err)
        elif (args.command == 'purge'):
            if not args.yes:
                yes = {'yes', 'y'}
                msg = ("*********This action can not be undone**********\n"
                       "*  You are about to purge any Pystol resource  *\n"
                       "* part of the pystol namespace in the cluster. *\n"
                       "* This includes any deployment, previous runs, *\n"
                       "*  RBAC rules, and any other Pystol resource.  *\n"
                       "*********This action can not be undone**********\n")
                print(msg)
                choice = input(
                    "Write 'yes' and press 'enter' to proceed: \n")
                if choice in yes:
                    print(u"\U0001F480" +
                          " Purging Pystol from the cluster.")
                    print("  " + u"\U0001F4A3" +
                          " Removing resources...")
                    purge_pystol()
                    print(u"\u2728" + " Pystol was removed completely.")
                else:
                    print("Cancelling...")
            else:
                print(u"\U0001F480" + " Purging Pystol from the cluster.")
                print("  " + u"\U0001F4A3" + " Removing resources...")
                purge_pystol()
                print(u"\u2728" + " Pystol was removed completely.")
            exit()
        elif (args.command == 'list'):
            print(u"\U0001F4DA" + " Listing the deployed actions"
                                  " from the cluster.")
            list_actions()
            print(u"\U0001F4A1" + " For further information use:"
                                  " pystol get <action_name> [--debug]")
            exit()
        elif (args.command == 'show'):
            if args.action:
                show_action(args.action)
            else:
                show_actions()
            exit()
        elif (args.command == 'get'):
            print(u"\U0001F4E4" + " Getting the details from"
                                  " an specific action.")
            get_action(args.action, args.debug)
            exit()
        elif (args.command == 'deploy'):
            print(u"\U0001F680" + " Deploying Pystol in the cluster.")
            deploy_pystol()
            print(u"\U0001F3C4" + " Done! Pystol is now deployed.")
            print(u"\U0001F550" + " Wait a few seconds to have the"
                                  " deployments up and running.")
            exit()

    except KeyboardInterrupt:
        pass
    except Exception as err:
        raise RuntimeError('There is something wrong...' + err)

    while not t2_stop.is_set() or not t1_stop.is_set():
        pass



#!/usr/bin/env python

"""
Copyright 2019 Pystol (pystol.org).

Licensed under the Apache License, Version 2.0 (the "License"); you may
not use this file except in compliance with the License. You may obtain
a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""

import kubernetes
from kubernetes.client.rest import ApiException

from pystol import __version__
from pystol.operator import load_kubernetes_config

pystol_version = __version__


def purge_pystol(api_client=None):
    """
    Purge Pystol from the cluster.

    This is a main component of the input for the controller
    """
    load_kubernetes_config()
    v1 = kubernetes.client.CoreV1Api(api_client=api_client)

    name = 'pystol'
    pretty = 'true'
    orphan_dependents = True
    # propagation_policy = 'Foreground'
    propagation_policy = 'Background'
    body = kubernetes.client.V1DeleteOptions()
    try:
        v1.delete_namespace(name,
                            pretty=pretty,
                            orphan_dependents=orphan_dependents,
                            propagation_policy=propagation_policy,
                            body=body)
        print("    " + u"\U0001F9F9" + " Namespace removed.")
    except ApiException:
        print("    " + u"\u2757" + " Namespace removing warning.")
        print("       Can't remove it, maybe it's gone...")

    name = 'pystol-service'
    namespace = 'pystol'
    pretty = 'true'
    orphan_dependents = True
    propagation_policy = 'Background'
    body = kubernetes.client.V1DeleteOptions()
    try:
        v1.delete_namespaced_service(name,
                                     namespace=namespace,
                                     pretty=pretty,
                                     orphan_dependents=orphan_dependents,
                                     propagation_policy=propagation_policy,
                                     body=body)
        print("    " + u"\U0001F9F9" + " Service removed.")
    except ApiException:
        print("    " + u"\u2757" + " Service removing warning.")
        print("       Can't remove it, maybe it's gone...")

    name = 'pystol-config'
    namespace = 'pystol'
    pretty = 'true'
    orphan_dependents = True
    propagation_policy = 'Background'
    body = kubernetes.client.V1DeleteOptions()
    try:
        v1.delete_namespaced_config_map(name,
                                        namespace=namespace,
                                        pretty=pretty,
                                        orphan_dependents=orphan_dependents,
                                        propagation_policy=propagation_policy,
                                        body=body)
        print("    " + u"\U0001F9F9" + " Config map removed.")
    except ApiException:
        print("    " + u"\u2757" + " Config map removing warning.")
        print("       Can't remove it, maybe it's gone...")

    name = 'pystol'
    namespace = 'pystol'
    pretty = 'true'
    orphans = True
    propagation = 'Background'
    body = kubernetes.client.V1DeleteOptions()
    try:
        v1.delete_namespaced_service_account(name,
                                             namespace=namespace,
                                             pretty=pretty,
                                             orphan_dependents=orphans,
                                             propagation_policy=propagation,
                                             body=body)
        print("    " + u"\U0001F9F9" + " Service account removed.")
    except ApiException:
        print("    " + u"\u2757" + " Service account removing warning.")
        print("       Can't remove it, maybe it's gone...")

    rbac = kubernetes.client.RbacAuthorizationV1Api(
        api_client=api_client)
    name = 'pystol'
    pretty = 'true'
    orphan_dependents = True
    propagation_policy = 'Background'
    body = kubernetes.client.V1DeleteOptions()
    try:
        rbac.delete_cluster_role(name,
                                 pretty=pretty,
                                 orphan_dependents=orphan_dependents,
                                 propagation_policy=propagation_policy,
                                 body=body)
        print("    " + u"\U0001F9F9" + " Cluster role removed.")
    except ApiException:
        print("    " + u"\u2757" + " Cluster role removing warning.")
        print("       Can't remove it, maybe it's gone...")

    rbac = kubernetes.client.RbacAuthorizationV1Api(
        api_client=api_client)
    name = 'pystol'
    pretty = 'true'
    orphan_dependents = True
    propagation_policy = 'Background'
    body = kubernetes.client.V1DeleteOptions()
    try:
        rbac.delete_cluster_role_binding(name,
                                         pretty=pretty,
                                         orphan_dependents=orphan_dependents,
                                         propagation_policy=propagation_policy,
                                         body=body)
        print("    " + u"\U0001F9F9" +
              " Cluster role binding removed.")
    except ApiException:
        print("    " + u"\u2757" +
              " Cluster role binding removing warning.")
        print("       Can't remove it, maybe it's gone...")

    ext = kubernetes.client.ApiextensionsV1beta1Api(
        api_client=api_client)
    name = 'pystolactions.pystol.org'
    pretty = 'true'
    orphans = True
    propagation = 'Background'
    body = kubernetes.client.V1DeleteOptions()
    try:
        ext.delete_custom_resource_definition(name,
                                              pretty=pretty,
                                              orphan_dependents=orphans,
                                              propagation_policy=propagation,
                                              body=body)
        print("    " + u"\U0001F9F9" + " CRD removed.")
    except ApiException:
        print("    " + u"\u2757" + " CRD removing warning.")
        print("       Can't remove it, maybe it's gone...")



#!/usr/bin/env python

"""
Copyright 2019 Pystol (pystol.org).

Licensed under the Apache License, Version 2.0 (the "License"); you may
not use this file except in compliance with the License. You may obtain
a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""

import os

from jinja2 import Template

import kubernetes
from kubernetes.client.rest import ApiException
from kubernetes.utils.create_from_yaml import FailToCreateError

from pystol import __version__
from pystol.operator import load_kubernetes_config

import yaml

pystol_version = __version__


def deploy_pystol(api_client=None):
    """
    Install Pystol from Python.

    This is a main component of the input for the controller
    """
    load_kubernetes_config()
    v1 = kubernetes.client.CoreV1Api(api_client=api_client)
    deployment = kubernetes.client.AppsV1Api(api_client=api_client)
    rbac = kubernetes.client.RbacAuthorizationV1Api(
        api_client=api_client)

    if api_client == None:
        apicli = kubernetes.client.ApiClient()
    else:
        apicli = api_client

    with open(os.path.join(os.path.dirname(__file__),
                           "templates/latest/pystol.upstreamvalues.yaml")) as f:
        values = yaml.safe_load(f)

    with open(os.path.join(os.path.dirname(__file__),
                           "templates/latest/pystol.namespace.yaml")) as f:
        try:
            resp = v1.create_namespace(
                body=yaml.safe_load(f))
            print("  " + u"\U0001F4E6" + " Namespace created.")
            print("     '%s'" % resp.metadata.name)
        except ApiException:
            print("  " + u"\u2757" + " Namespace creation warning.")
            print("     Maybe it is already created.")

    '''
    with open(os.path.join(os.path.dirname(__file__),
                           "templates/latest/pystol.configmap.yaml.j2")) as f:
        template = Template(f.read())
        cm = template.render(values)
        try:
            resp = v1.create_namespaced_config_map(
                body=yaml.safe_load(cm), namespace="pystol")
            print("  " + u"\U0001F4E6" + " Config map created.")
            print("     '%s'" % resp.metadata.name)
        except ApiException:
            print("  " + u"\u2757" + " Config map creation warning.")
            print("     Maybe it is already created.")
    '''
    try:
        resp = kubernetes.utils.create_from_yaml(
            k8s_client=apicli,
            yaml_file=os.path.join(os.path.dirname(__file__),
                                   "templates/latest/pystol.crd.yaml"),
            namespace="pystol"
        )
        print("  " + u"\U0001F4E6" + " CRD created.")
        print("     '%s'" % resp.metadata.name)
    except FailToCreateError:
        print("  " + u"\u2757" + " CRD creation warning.")
        print("     Maybe it is already created.")
    except Exception:
        print("  " + u"\U0001F4E6" + " CRD created.")
        print("     We need to wait for a permanent fix, until then...")
        print("     https://github.com/kubernetes-client/python/issues/1022")
        # print("CRD problem - ApiClient->create_from_yaml: %s\n" % e)
        # print("The CRD was created but an exception is raised")
        # print("Other error, see:")
        # print("https://github.com/kubernetes-client/python/issues/1022")

    with open(os.path.join(os.path.dirname(__file__),
                           "templates/latest/pystol.serviceaccount.yaml")) as f:
        try:
            resp = v1.create_namespaced_service_account(
                namespace="pystol",
                body=yaml.safe_load(f))
            print("  " + u"\U0001F4E6" + " Service account created.")
            print("     '%s'" % resp.metadata.name)
        except ApiException:
            print("  " + u"\u2757" +
                  " Service account creation warning.")
            print("     Maybe it is already created.")

    with open(os.path.join(os.path.dirname(__file__),
                           "templates/latest/pystol.clusterrole.yaml")) as f:
        try:
            resp = rbac.create_cluster_role(
                body=yaml.safe_load(f))
            print("  " + u"\U0001F4E6" + " Role created.")
            print("     '%s'" % resp.metadata.name)
        except ApiException:
            print("  " + u"\u2757" + " Role creation warning.")
            print("     Maybe it is already created.")

    with open(os.path.join(os.path.dirname(__file__),
                           "templates/latest/pystol.clusterrolebinding.yaml")) as f:
        try:
            resp = rbac.create_cluster_role_binding(
                body=yaml.safe_load(f))
            print("  " + u"\U0001F4E6" +
                  " Cluster role bindings created.")
            print("     '%s'"
                  % resp.metadata.name)
        except ApiException:
            print("  " + u"\u2757"
                       + " Cluster role binding creation warning.")
            print("     Maybe it is already created.")

    with open(os.path.join(os.path.dirname(__file__),
                           "templates/latest/pystol.controller.yaml.j2")) as f:
        template = Template(f.read())
        rendered_deployment = template.render(values)
        try:
            resp = deployment.create_namespaced_deployment(
                body=yaml.safe_load(rendered_deployment),
                namespace="pystol")
            print("  " + u"\U0001F4E6"
                       + " Operator deployment created.")
            print("     '%s'" % resp.metadata.name)
        except ApiException:
            print("  " + u"\u2757"
                       + " Operator deployment creation warning.")
            print("     Maybe it is already created.")

    with open(os.path.join(os.path.dirname(__file__),
                           "templates/latest/pystol.ui.yaml.j2")) as f:
        template = Template(f.read())
        rendered_deployment = template.render(values)
        try:
            resp = deployment.create_namespaced_deployment(
                body=yaml.safe_load(rendered_deployment),
                namespace="pystol")
            print("  " + u"\U0001F4E6" + " UI operator created.")
            print("     '%s'"
                  % resp.metadata.name)
        except ApiException:
            print("  " + u"\u2757" + " UI deployment creation warning.")
            print("     Maybe it is already created.")

    with open(os.path.join(os.path.dirname(__file__),
                           "templates/latest/pystol.service.yaml")) as f:
        try:
            resp = v1.create_namespaced_service(
                namespace="pystol",
                body=yaml.safe_load(f))
            print("  " + u"\U0001F4E6" + " Service created.")
            print("     '%s'" % resp.metadata.name)
        except ApiException:
            print("  " + u"\u2757" + " Service creation warning.")
            print("     Maybe it is already created.")



#!/usr/bin/env python

"""
Copyright 2019 Pystol (pystol.org).

Licensed under the Apache License, Version 2.0 (the "License"); you may
not use this file except in compliance with the License. You may obtain
a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""

import os
import tempfile

# Main logging file
PYSTOL_LOG_FOLDER = tempfile.gettempdir()
PYSTOL_LOG_FILE = os.path.join(PYSTOL_LOG_FOLDER, '/tmp/pystol.log')

# Current branch
PYSTOL_BRANCH = 'master'

# CRD Settings
CRD_DOMAIN = 'pystol.org'
CRD_GROUP = 'pystol.org'
CRD_VERSION = 'v1alpha1'
CRD_PLURAL = 'pystolactions'
CRD_NAMESPACE = 'pystol'

# Type methods maps
LIST_TYPES_MAP = {
    'configmap': 'list_namespaced_config_map',
    'secret': 'list_namespaced_secret',
}

CREATE_TYPES_MAP = {
    'configmap': 'create_namespaced_config_map',
    'secret': 'create_namespaced_secret',
}

# Allowed events
ALLOWED_EVENT_TYPES = {'ADDED', 'UPDATED'}



#!/usr/bin/env python

"""
Copyright 2019 Pystol (pystol.org).

Licensed under the Apache License, Version 2.0 (the "License"); you may
not use this file except in compliance with the License. You may obtain
a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""

import os
from sys import version_info

from setuptools import find_packages, setup

if version_info < (3, 5):
    raise RuntimeError(
        'Python 3.5 or greater is required'
    )

_NAME = 'pystol'
_DESCRIPTION = 'The Pystol CLI'
_REVISION = '0.8.17'

pystol_revision = os.environ.get('PYSTOL_REVISION', "")
if (pystol_revision != ""):
    _REVISION = _REVISION + "." + pystol_revision

if os.path.isfile('../README.md'):
    with open('../README.md') as f:
        long_description = f.read()
else:
    long_description = _DESCRIPTION

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name=_NAME,
    version=_REVISION,
    description=_DESCRIPTION,
    long_description_content_type='text/markdown',
    long_description=long_description,
    url='https://www.pystol.org',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Testing',
    ],
    author='Carlos Camacho',
    author_email='carloscamachoucv@gmail.com',
    include_package_data=True,
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            '{0} = {0}.cli:main'.format(_NAME),
        ]
    }
)



#!/usr/bin/env python

"""
Copyright 2019 Pystol (pystol.org).

Licensed under the Apache License, Version 2.0 (the "License"); you may
not use this file except in compliance with the License. You may obtain
a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""

import logging
import sys
from logging.handlers import TimedRotatingFileHandler

from pystol.const import PYSTOL_LOG_FILE

FORMATTER = logging.Formatter("%(asctime)s - %(name)s - "
                              "%(levelname)s - %(message)s")


def get_console_handler():
    """
    Get logger console handler.

    This is a main component of the input for the controller
    """
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler


def get_file_handler():
    """
    Get logger file handler.

    This is a main component of the input for the controller
    """
    file_handler = TimedRotatingFileHandler(
        PYSTOL_LOG_FILE, when='midnight')
    file_handler.setFormatter(FORMATTER)
    return file_handler


def get_logger(logger_name):
    """
    Get logger.

    This is a main component of the input for the controller
    """
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    # We do not want to print in the console
    # logger.addHandler(get_console_handler())
    logger.addHandler(get_file_handler())
    logger.propagate = False
    return logger



#!/usr/bin/env python

"""
Copyright 2019 Pystol (pystol.org).

Licensed under the Apache License, Version 2.0 (the "License"); you may
not use this file except in compliance with the License. You may obtain
a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""

import json
import os
import random
import string
import sys

import kubernetes
from kubernetes.client.rest import ApiException

from prettytable import PrettyTable

from pystol import __version__
from pystol.const \
    import CRD_DOMAIN, CRD_NAMESPACE, CRD_PLURAL, CRD_VERSION
from pystol.logger import get_logger

pystol_version = __version__

# custom_obj = kubernetes.client.CustomObjectsApi()
# v1 = kubernetes.client.CoreV1Api()

#
# We load the Kubernetes cluster config depending
# where we execute the operator from.
#


def load_kubernetes_config():
    """
    Load the initial config details.

    We load the config depending where we execute the code from
    """
    try:
        if 'KUBERNETES_PORT' in os.environ:
            # We set up the client from within a k8s pod
            kubernetes.config.load_incluster_config()
        elif 'KUBECONFIG' in os.environ:
            kubernetes.config.load_kube_config(
                os.getenv('KUBECONFIG'))
        else:
            kubernetes.config.load_kube_config()
    except Exception as e:
        message = ("---\n"
                   "The Python Kubernetes client could not be configured "
                   "at this time.\n"
                   "You need a working Kubernetes deployment to make "
                   "Pystol work.\n"
                   "Check the following:\n"
                   "Use the env var KUBECONFIG with the path to your K8s "
                   "config file like:\n"
                   "    export KUBECONFIG=~/.kube/config\n"
                   "Or run Pystol from within the cluster to make use of "
                   "load_incluster_config.\n"
                   "Version: %s"
                   "Error: %s" % (pystol_version, e))

#
# Part of the operation in charge of adding the custom resources
# to the Kubernetes cluster, this will create an object with
# the CLI parameters.
#


def insert_pystol_object(namespace,
                         collection,
                         role,
                         source,
                         extra_vars,
                         api_client=None):
    """
    Determine where we will insert the CR.

    This is a main component of the input for the controller
    """
    load_kubernetes_config()
    custom_obj = kubernetes.client.CustomObjectsApi(
        api_client=api_client)

    resource = {
        "apiVersion": CRD_DOMAIN + "/" + CRD_VERSION,
        "kind": "PystolAction",
        "metadata": {"name": "pystol-action-" +
                             namespace +
                             "-" +
                             collection +
                             "-" +
                             role +
                             "-" +
                             id_generator()},
        "spec": {"namespace": namespace,
                 "collection": collection,
                 "role": role,
                 "source": source,
                 "extra_vars": extra_vars,
                 "action_state": "PystolActionCreating",
                 "workflow_state": "PystolOperatorWaitingAction",
                 "action_stderr": "{}",
                 "action_stdout": "{}"},
    }

    try:
        # create the resource
        api_response = custom_obj.create_namespaced_custom_object(
            group=CRD_DOMAIN,
            version=CRD_VERSION,
            namespace=CRD_NAMESPACE,
            plural=CRD_PLURAL,
            body=resource,
        )
    except ApiException as e:
        get_logger("insert_pystol_object").debug(e)
        return False

    x = PrettyTable()
    x.field_names = ["Name",
                     "Creation",
                     "Action state",
                     "Workflow state"]

    x.add_row([api_response['metadata']['name'],
               api_response['metadata']['creationTimestamp'],
               api_response['spec']['action_state'],
               api_response['spec']['workflow_state']])

    print(x)
    return True

#
# Part of the operator which will handle the process of new
# PystolAction objects added in the cluster, It will watch
# for a new object and create a software deployment to run
# the selected Pystol action.
#


def watch_for_pystol_timeouts(stop,
                              api_client=None):
    """
    Watch for action with timeouts.

    This method will listen for custom objects
    that times out.
    """
    while True:
        return True


def watch_for_pystol_objects(stop,
                             api_client=None):
    """
    Initiate the main listening method.

    This method will listen for custom objects
    added to the cluster.
    The watcher is defined here.
    """
    load_kubernetes_config()
    custom_obj = kubernetes.client.CustomObjectsApi(
        api_client=api_client)

    w = kubernetes.watch.Watch()
    for event in w.stream(custom_obj.list_cluster_custom_object,
                          CRD_DOMAIN,
                          CRD_VERSION,
                          CRD_PLURAL,
                          resource_version=''):
        obj = event["object"]
        operation = event['type']
        spec = obj.get("spec")
        if not spec:
            continue
        metadata = obj.get("metadata")
        name = metadata['name']
        done = spec.get("executed", False)
        if done:
            continue
        print("Processing %s on %s" % (operation, name))
        execute_pystol_action(obj)


def execute_pystol_action(obj,
                          api_client=None):
    """
    Execute the Pystol action.

    This method will execute the Pystol action
    defined in the custom object.
    """
    load_kubernetes_config()
    custom_obj = kubernetes.client.CustomObjectsApi(
        api_client=api_client)

    metadata = obj.get("metadata")
    if not metadata:
        print("No metadata in object, skipping: %s"
              % json.dumps(obj, indent=1))
        return
    name = metadata.get("name")
    namespace = metadata.get("namespace")

    # The main Pystol object initial info are the parameters:
    # These values should be the ones defined in the CRD.

    action_spec_params = obj.get("spec")

    action_namespace = action_spec_params["namespace"]
    action_collection = action_spec_params["collection"]
    action_role = action_spec_params["role"]
    action_source = action_spec_params["source"]
    action_extra_vars = action_spec_params["extra_vars"]
    action_action_state = action_spec_params["action_state"]
    action_workflow_state = action_spec_params["workflow_state"]
    action_action_stdout = action_spec_params["action_stdout"]
    action_action_stderr = action_spec_params["action_stderr"]

    # TODO: get object here to update it
    updt = custom_obj.get_namespaced_custom_object(group=CRD_DOMAIN,
                                                   version=CRD_VERSION,
                                                   namespace=namespace,
                                                   plural=CRD_PLURAL,
                                                   name=name)
    updt["spec"]["executed"] = True
    updt["spec"]["workflow_state"] = "PystolOperatorStartProcessingAction"
    # Processing action
    print("Updating processing action: %s" % name)
    custom_obj.patch_namespaced_custom_object(CRD_DOMAIN,
                                              CRD_VERSION,
                                              namespace,
                                              CRD_PLURAL,
                                              name,
                                              updt)

    # Create the job definition
    api_instance = kubernetes.client.BatchV1Api(api_client=api_client)

    # TODO: This should be configurable
    container_image = "quay.io/pystol/pystol-operator-stable:latest"

    body = kube_create_job_object(name=name,
                                  container_image=container_image,
                                  namespace=namespace,
                                  env_vars={"VAR": "TESTING"},
                                  # CRD variables
                                  action_namespace=action_namespace,
                                  action_collection=action_collection,
                                  action_role=action_role,
                                  action_source=action_source,
                                  action_extra_vars=action_extra_vars,
                                  action_action_state=action_action_state,
                                  action_workflow_state=action_workflow_state,
                                  action_action_stdout=action_action_stdout,
                                  action_action_stderr=action_action_stderr)

    try:
        api_response = api_instance.create_namespaced_job("pystol",
                                                          body,
                                                          pretty=True)
        print(api_response)
    except ApiException as e:
        print("Exception when calling BatchV1Api->create_namespaced_job: %s\n"
              % e)

    # Updating the CR with - Creating job
    print("Updating creating job: %s" % name)
    # TODO: get object here to update it
    updt = custom_obj.get_namespaced_custom_object(group=CRD_DOMAIN,
                                                   version=CRD_VERSION,
                                                   namespace=namespace,
                                                   plural=CRD_PLURAL,
                                                   name=name)
    updt["spec"]["workflow_state"] = "PystolOperatorCreatingJob"
    # Processing action
    print("Updating processing action: %s" % name)
    custom_obj.patch_namespaced_custom_object(CRD_DOMAIN,
                                              CRD_VERSION,
                                              namespace,
                                              CRD_PLURAL,
                                              name,
                                              updt)

    return


def kube_create_job_object(name,
                           container_image,
                           namespace,
                           env_vars,
                           action_namespace,
                           action_collection,
                           action_role,
                           action_source,
                           action_extra_vars,
                           action_action_state,
                           action_workflow_state,
                           action_action_stdout,
                           action_action_stderr):
    """
    Create the Pystol job.

    This method will create the k8s job
    that launchs the action.
    """
    load_kubernetes_config()

    # Body is the object Body
    body = kubernetes.client.V1Job(api_version="batch/v1",
                                   kind="Job")
    # Body needs Metadata
    # Attention: Each JOB must have a different name!
    body.metadata = kubernetes.client.V1ObjectMeta(namespace=namespace,
                                                   name=name)
    # And a Status
    body.status = kubernetes.client.V1JobStatus()
    # Now we start with the Template...
    template = kubernetes.client.V1PodTemplate()
    template.template = kubernetes.client.V1PodTemplateSpec()
    # Passing Arguments in Env:
    env_list = []
    for env_name, env_value in env_vars.items():
        env_list.append(kubernetes.client.V1EnvVar(name=env_name,
                                                   value=env_value))

    # Python interpreter as an extra variable
    # python object to be appended
    y = {"ansible_python_interpreter": "/usr/bin/python3",
         "pystol_action_id": name}
    # parsing JSON string:
    extra_ansible_vars = json.loads(action_extra_vars)
    # appending the data
    extra_ansible_vars.update(y)

    # Recovery ansible variables for the log role
    # In the case a user calls an unrecognized action
    # the container will fail and we will not be able to log this
    # failure.

    # We can use the same extra_ansible_vars as the workflow and
    # action state should not be used by any action from the
    # operator, they are calculated as long as the collection
    # is executed.

    y2 = {"ansible_python_interpreter":
          "/usr/bin/python3",
          "pystol_action_id":
          name,
          "pystol_patch_workflow_state":
          "PystolOperatorEnded",
          "pystol_patch_action_state":
          "PystolActionEndedFail",
          "pystol_patch_action_stdout":
          "This-action-did-not-finish-correctly",
          "pystol_patch_action_stderr":
          "Probably-the-action-was-not-found-Check-the-logs"}
    # rec_extra_ansible_vars = json.loads("{}")
    # rec_extra_ansible_vars.update(y2)
    extra_ansible_vars.update(y2)

    command = ["/bin/bash"]

    if (action_source == '' or action_source is None):
        action_source = "galaxy.ansible.com"

    if (action_source == "galaxy.ansible.com"):
        args = ["-c",
                "echo '---' > req.yml; \
                 echo 'collections:' >> req.yml; \
                 echo '- name: " + action_namespace + "." + action_collection + "' >> req.yml; \
                 echo '  source: https://" + action_source + "' >> req.yml; \
                 ansible-galaxy collection install --force -r req.yml; \
                 ansible -m include_role \
                   -a 'name=" + action_namespace + "." + action_collection + "." + action_role + "' \
                   -e '" + str(extra_ansible_vars) + "' localhost -vv || \
                 ansible -m include_role -a 'name=pystol.actions.patch' -e '" + str(extra_ansible_vars) + "' localhost -vv; \
                 exit 0"]
    else:
        args = ["-c",
                "echo '---'; \
                 git clone " + action_source + " cloned_repo; \
                 cd cloned_repo; \
                 cd " + action_collection + "; \
                 mkdir -p releases; \
                 ansible-galaxy collection build -v \
                                                 --force \
                                                 --output-path releases/; \
                 cd releases; \
                 LATEST=$(ls *.tar.gz | grep -v latest | sort -V | tail -n1); \
                 ansible-galaxy collection install --force $LATEST; \
                 ansible -m include_role \
                   -a 'name=" + action_namespace + "." + action_collection + "." + action_role + "' \
                   -e '" + str(extra_ansible_vars) + "' localhost -vv || \
                 ansible -m include_role -a 'name=pystol.actions.patch' -e '" + str(extra_ansible_vars) + "' localhost -vv; \
                 exit 0"]

    container = kubernetes.client.V1Container(name=name,
                                              image=container_image,
                                              command=command,
                                              args=args,
                                              env=env_list)
    template.template.spec = kubernetes.client.V1PodSpec(
        containers=[container],
        restart_policy='Never')
    # And finaly we can create our V1JobSpec!
    body.spec = kubernetes.client.V1JobSpec(ttl_seconds_after_finished=600,
                                            template=template.template)
    return body


def id_generator(size=5, chars=string.ascii_lowercase + string.digits):
    """
    Generate a random sufix.

    This method will generate a
    random sufix for the created resources.
    """
    return ''.join(random.choice(chars) for _ in range(size))
